==> sqlsaber-0.32.0.tar.gz: sqlsaber-0.32.0/tests/test_database/test_duckdb_module.py <==
"""Comprehensive tests for DuckDB database module."""

import duckdb
import pytest

from sqlsaber.database.duckdb import DuckDBConnection, DuckDBSchemaIntrospector


class TestDuckDBConnection:
    """Test DuckDB connection functionality."""

    def test_connection_string_parsing(self):
        """Test DuckDB connection string parsing variations."""
        # Test full duckdb:// format
        conn1 = DuckDBConnection("duckdb:///path/to/database.duckdb")
        assert conn1.database_path == "path/to/database.duckdb"

        # Test short duckdb: format
        conn2 = DuckDBConnection("duckdb://database.duckdb")
        assert conn2.database_path == "database.duckdb"

        # Test plain path
        conn3 = DuckDBConnection("database.duckdb")
        assert conn3.database_path == "database.duckdb"

        # Test in-memory
        conn4 = DuckDBConnection("")
        assert conn4.database_path == ":memory:"

    @pytest.mark.asyncio
    async def test_memory_database_operations(self):
        """Test DuckDB in-memory database operations."""
        conn = DuckDBConnection(":memory:")

        # Test basic query
        result = await conn.execute_query("SELECT 42 as answer")
        assert len(result) == 1
        assert result[0]["answer"] == 42

        # Test table creation (each query runs in separate rolled-back transaction)
        await conn.execute_query("CREATE TABLE test (id INTEGER, name VARCHAR)")

        # Due to transaction rollback, subsequent queries won't see the table
        # But the creation itself should succeed within its transaction

        await conn.close()

    @pytest.mark.asyncio
    async def test_file_database_operations(self, tmp_path):
        """Test DuckDB file database operations."""
        db_path = tmp_path / "test.duckdb"
        conn = DuckDBConnection(str(db_path))

        # Create table with DuckDB-specific features
        await conn.execute_query("""
            CREATE TABLE products (
                id INTEGER PRIMARY KEY,
                name VARCHAR(100) NOT NULL,
                price DECIMAL(10,2) DEFAULT 0.00,
                tags VARCHAR[],  -- DuckDB array type
                metadata JSON,   -- DuckDB JSON type
                created_at TIMESTAMP DEFAULT NOW()
            )
        """)

        # Test DuckDB-specific data types
        result = await conn.execute_query("SELECT 1 as id, 'Test' as name")
        assert result[0]["name"] == "Test"

        await conn.close()

    @pytest.mark.asyncio
    async def test_parameterized_queries(self):
        """Test DuckDB parameterized queries."""
        conn = DuckDBConnection(":memory:")

        await conn.execute_query("CREATE TABLE params_test (id INTEGER, value VARCHAR)")

        # Test that parameterized queries are accepted (syntax test)
        try:
            await conn.execute_query(
                "INSERT INTO params_test VALUES (?, ?)", 1, "test_value"
            )
        except Exception:
            pass  # Expected due to transaction rollback, but parameter syntax was tested

        await conn.close()

    @pytest.mark.asyncio
    async def test_complex_queries(self):
        """Test DuckDB complex query capabilities."""
        conn = DuckDBConnection(":memory:")

        # Test array operations
        result = await conn.execute_query(
            "SELECT [1, 2, 3] as arr, 'hello' as greeting"
        )
        assert result[0]["greeting"] == "hello"

        # Test JSON operations
        result = await conn.execute_query(
            "SELECT {'name': 'Alice', 'age': 30} as person"
        )
        assert len(result) == 1

        await conn.close()

    @pytest.mark.asyncio
    async def test_error_handling(self):
        """Test DuckDB error handling."""
        conn = DuckDBConnection(":memory:")

        # Test syntax error
        with pytest.raises(Exception):
            await conn.execute_query("INVALID SQL SYNTAX")

        # Test table not found
        with pytest.raises(Exception):
            await conn.execute_query("SELECT * FROM nonexistent_table")

        await conn.close()


class TestDuckDBSchemaIntrospector:
    """Test DuckDB schema introspection functionality."""

    @pytest.mark.asyncio
    async def test_table_listing(self, tmp_path):
        """Test DuckDB table listing."""
        db_path = tmp_path / "schema_test.duckdb"

        # Create database with tables and views
        with duckdb.connect(str(db_path)) as db_conn:
            db_conn.execute("CREATE TABLE customers (id INTEGER, name VARCHAR)")
            db_conn.execute("CREATE TABLE orders (id INTEGER, customer_id INTEGER)")
            db_conn.execute(
                "CREATE VIEW customer_orders AS SELECT * FROM customers JOIN orders ON customers.id = orders.customer_id"
            )

        # Test introspection
        conn = DuckDBConnection(str(db_path))
        introspector = DuckDBSchemaIntrospector()

        tables = await introspector.list_tables_info(conn)

        table_names = {table["table_name"] for table in tables}
        table_types = {table["table_name"]: table["table_type"] for table in tables}

        assert "customers" in table_names
        assert "orders" in table_names
        assert "customer_orders" in table_names

        # DuckDB uses different type names
        assert table_types["customers"] == "BASE TABLE"
        assert table_types["orders"] == "BASE TABLE"
        assert table_types["customer_orders"] == "VIEW"

        await conn.close()

    @pytest.mark.asyncio
    async def test_column_introspection_with_duckdb_types(self, tmp_path):
        """Test DuckDB column introspection with DuckDB-specific types."""
        db_path = tmp_path / "types_test.duckdb"

        with duckdb.connect(str(db_path)) as db_conn:
            db_conn.execute("""
                CREATE TABLE duckdb_types (
                    id INTEGER PRIMARY KEY,
                    name VARCHAR(100) DEFAULT 'Unknown',
                    price DECIMAL(10,2) DEFAULT 0.00,
                    active BOOLEAN DEFAULT TRUE,
                    tags VARCHAR[] DEFAULT [],
                    metadata JSON,
                    created_at TIMESTAMP DEFAULT NOW(),
                    updated_at TIMESTAMPTZ DEFAULT NOW(),
                    data BLOB,
                    rating DOUBLE DEFAULT 0.0
                )
            """)

        conn = DuckDBConnection(str(db_path))
        introspector = DuckDBSchemaIntrospector()

        tables = await introspector.get_tables_info(conn)
        columns = await introspector.get_columns_info(conn, tables)

        # Organize columns for easier testing
        column_info = {
            col["column_name"]: col
            for col in columns
            if col["table_name"] == "duckdb_types"
        }

        # Check DuckDB-specific types
        assert column_info["name"]["data_type"] == "VARCHAR"
        assert column_info["price"]["data_type"] == "DECIMAL(10,2)"
        assert column_info["active"]["data_type"] == "BOOLEAN"
        assert column_info["created_at"]["data_type"] == "TIMESTAMP"
        assert column_info["rating"]["data_type"] == "DOUBLE"

        # Check defaults
        assert column_info["name"]["column_default"] is not None
        assert column_info["price"]["column_default"] is not None
        assert column_info["active"]["column_default"] is not None

        await conn.close()

    @pytest.mark.asyncio
    async def test_primary_key_detection(self, tmp_path):
        """Test DuckDB primary key detection."""
        db_path = tmp_path / "pk_test.duckdb"

        with duckdb.connect(str(db_path)) as db_conn:
            db_conn.execute(
                "CREATE TABLE single_pk (id INTEGER PRIMARY KEY, data VARCHAR)"
            )
            db_conn.execute(
                "CREATE TABLE composite_pk (user_id INTEGER, role_id INTEGER, data VARCHAR, PRIMARY KEY (user_id, role_id))"
            )

        conn = DuckDBConnection(str(db_path))
        introspector = DuckDBSchemaIntrospector()

        tables = await introspector.get_tables_info(conn)
        primary_keys = await introspector.get_primary_keys_info(conn, tables)

        pk_by_table = {}
        for pk in primary_keys:
            table_name = pk["table_name"]
            if table_name not in pk_by_table:
                pk_by_table[table_name] = []
            pk_by_table[table_name].append(pk["column_name"])

        assert "id" in pk_by_table["single_pk"]
        assert "user_id" in pk_by_table["composite_pk"]
        assert "role_id" in pk_by_table["composite_pk"]

        await conn.close()

    @pytest.mark.asyncio
    async def test_foreign_key_detection(self, tmp_path):
        """Test DuckDB foreign key detection."""
        db_path = tmp_path / "fk_test.duckdb"

        with duckdb.connect(str(db_path)) as db_conn:
            db_conn.execute(
                "CREATE TABLE departments (id INTEGER PRIMARY KEY, name VARCHAR)"
            )
            db_conn.execute("""
                CREATE TABLE employees (
                    id INTEGER PRIMARY KEY,
                    name VARCHAR,
                    dept_id INTEGER,
                    FOREIGN KEY (dept_id) REFERENCES departments(id)
                )
            """)

        conn = DuckDBConnection(str(db_path))
        introspector = DuckDBSchemaIntrospector()

        tables = await introspector.get_tables_info(conn)
        foreign_keys = await introspector.get_foreign_keys_info(conn, tables)

        # Check foreign key relationship
        fk_found = False
        for fk in foreign_keys:
            if (
                fk["table_name"] == "employees"
                and fk["column_name"] == "dept_id"
                and fk["foreign_table_name"] == "departments"
                and fk["foreign_column_name"] == "id"
            ):
                fk_found = True
                break

        assert fk_found, "Foreign key relationship not detected"
        await conn.close()

    @pytest.mark.asyncio
    async def test_index_detection(self, tmp_path):
        """Test DuckDB index detection."""
        db_path = tmp_path / "index_test.duckdb"

        with duckdb.connect(str(db_path)) as db_conn:
            db_conn.execute(
                "CREATE TABLE items (id INTEGER, name VARCHAR, category VARCHAR)"
            )
            db_conn.execute("CREATE INDEX idx_item_name ON items(name)")
            db_conn.execute("CREATE UNIQUE INDEX idx_item_id ON items(id)")

        conn = DuckDBConnection(str(db_path))
        introspector = DuckDBSchemaIntrospector()

        tables = await introspector.get_tables_info(conn)
        indexes = await introspector.get_indexes_info(conn, tables)

        index_names = [idx["index_name"] for idx in indexes]

        assert "idx_item_name" in index_names
        assert "idx_item_id" in index_names

        # Check index properties
        for idx in indexes:
            if idx["index_name"] == "idx_item_name":
                assert not idx["is_unique"]
                assert "name" in idx["column_names"]
            elif idx["index_name"] == "idx_item_id":
                assert idx["is_unique"]
                assert "id" in idx["column_names"]

        await conn.close()

    @pytest.mark.asyncio
    async def test_table_pattern_filtering(self, tmp_path):
        """Test DuckDB table pattern filtering."""
        db_path = tmp_path / "pattern_test.duckdb"

        with duckdb.connect(str(db_path)) as db_conn:
            db_conn.execute("CREATE TABLE log_events (id INTEGER)")
            db_conn.execute("CREATE TABLE log_errors (id INTEGER)")
            db_conn.execute("CREATE TABLE user_data (id INTEGER)")
            db_conn.execute("CREATE TABLE config_settings (id INTEGER)")

        conn = DuckDBConnection(str(db_path))
        introspector = DuckDBSchemaIntrospector()

        # Test pattern matching
        log_tables = await introspector.get_tables_info(conn, "log_%")
        log_table_names = {table["table_name"] for table in log_tables}

        assert "log_events" in log_table_names
        assert "log_errors" in log_table_names
        assert "user_data" not in log_table_names
        assert "config_settings" not in log_table_names

        await conn.close()

    @pytest.mark.asyncio
    async def test_schema_qualified_tables(self, tmp_path):
        """Test DuckDB schema-qualified table operations."""
        db_path = tmp_path / "schema_test.duckdb"

        with duckdb.connect(str(db_path)) as db_conn:
            # Create custom schema
            db_conn.execute("CREATE SCHEMA analytics")
            db_conn.execute("CREATE TABLE analytics.reports (id INTEGER, name VARCHAR)")
            db_conn.execute("CREATE TABLE main.users (id INTEGER, name VARCHAR)")

        conn = DuckDBConnection(str(db_path))
        introspector = DuckDBSchemaIntrospector()

        # Test schema.table pattern
        analytics_tables = await introspector.get_tables_info(conn, "analytics.%")
        analytics_table_names = {
            f"{table['table_schema']}.{table['table_name']}"
            for table in analytics_tables
        }

        assert "analytics.reports" in analytics_table_names
        assert "main.users" not in analytics_table_names

        await conn.close()

    @pytest.mark.asyncio
    async def test_duckdb_specific_features(self, tmp_path):
        """Test DuckDB-specific SQL features in introspection."""
        db_path = tmp_path / "advanced.duckdb"

        with duckdb.connect(str(db_path)) as db_conn:
            db_conn.execute("""
                CREATE TABLE advanced_features (
                    id INTEGER PRIMARY KEY,
                    json_data JSON,
                    array_data INTEGER[],
                    struct_data STRUCT(name VARCHAR, age INTEGER)
                )
            """)

        # Now test introspection
        conn = DuckDBConnection(str(db_path))
        introspector = DuckDBSchemaIntrospector()

        tables = await introspector.get_tables_info(conn)
        columns = await introspector.get_columns_info(conn, tables)

        # Check advanced data types are handled
        column_names = [col["column_name"] for col in columns]

        assert "id" in column_names
        assert "json_data" in column_names
        assert "array_data" in column_names

        await conn.close()

    @pytest.mark.asyncio
    async def test_empty_database_handling(self):
        """Test handling of empty DuckDB database."""
        conn = DuckDBConnection(":memory:")
        introspector = DuckDBSchemaIntrospector()

        tables = await introspector.list_tables_info(conn)
        assert len(tables) == 0

        columns = await introspector.get_columns_info(conn, [])
        assert len(columns) == 0

        foreign_keys = await introspector.get_foreign_keys_info(conn, [])
        assert len(foreign_keys) == 0

        await conn.close()

==> sqlsaber-0.37.0.tar.gz: sqlsaber-0.37.0/tests/test_database/test_duckdb_module.py <==
"""Comprehensive tests for DuckDB database module."""

import duckdb
import pytest

from sqlsaber.database.duckdb import DuckDBConnection, DuckDBSchemaIntrospector


class TestDuckDBConnection:
    """Test DuckDB connection functionality."""

    def test_connection_string_parsing(self):
        """Test DuckDB connection string parsing variations."""
        # Test full duckdb:// format
        conn1 = DuckDBConnection("duckdb:///path/to/database.duckdb")
        assert conn1.database_path == "path/to/database.duckdb"

        # Test short duckdb: format
        conn2 = DuckDBConnection("duckdb://database.duckdb")
        assert conn2.database_path == "database.duckdb"

        # Test plain path
        conn3 = DuckDBConnection("database.duckdb")
        assert conn3.database_path == "database.duckdb"

        # Test in-memory
        conn4 = DuckDBConnection("")
        assert conn4.database_path == ":memory:"

    @pytest.mark.asyncio
    async def test_memory_database_operations(self):
        """Test DuckDB in-memory database operations."""
        conn = DuckDBConnection(":memory:")

        # Test basic query
        result = await conn.execute_query("SELECT 42 as answer")
        assert len(result) == 1
        assert result[0]["answer"] == 42

        # Test table creation (each query runs in separate rolled-back transaction)
        await conn.execute_query("CREATE TABLE test (id INTEGER, name VARCHAR)")

        # Due to transaction rollback, subsequent queries won't see the table
        # But the creation itself should succeed within its transaction

        await conn.close()

    @pytest.mark.asyncio
    async def test_file_database_operations(self, tmp_path):
        """Test DuckDB file database operations."""
        db_path = tmp_path / "test.duckdb"
        conn = DuckDBConnection(str(db_path))

        # Create table with DuckDB-specific features
        await conn.execute_query("""
            CREATE TABLE products (
                id INTEGER PRIMARY KEY,
                name VARCHAR(100) NOT NULL,
                price DECIMAL(10,2) DEFAULT 0.00,
                tags VARCHAR[],  -- DuckDB array type
                metadata JSON,   -- DuckDB JSON type
                created_at TIMESTAMP DEFAULT NOW()
            )
        """)

        # Test DuckDB-specific data types
        result = await conn.execute_query("SELECT 1 as id, 'Test' as name")
        assert result[0]["name"] == "Test"

        await conn.close()

    @pytest.mark.asyncio
    async def test_parameterized_queries(self):
        """Test DuckDB parameterized queries."""
        conn = DuckDBConnection(":memory:")

        await conn.execute_query("CREATE TABLE params_test (id INTEGER, value VARCHAR)")

        # Test that parameterized queries are accepted (syntax test)
        try:
            await conn.execute_query(
                "INSERT INTO params_test VALUES (?, ?)", 1, "test_value"
            )
        except Exception:
            pass  # Expected due to transaction rollback, but parameter syntax was tested

        await conn.close()

    @pytest.mark.asyncio
    async def test_complex_queries(self):
        """Test DuckDB complex query capabilities."""
        conn = DuckDBConnection(":memory:")

        # Test array operations
        result = await conn.execute_query(
            "SELECT [1, 2, 3] as arr, 'hello' as greeting"
        )
        assert result[0]["greeting"] == "hello"

        # Test JSON operations
        result = await conn.execute_query(
            "SELECT {'name': 'Alice', 'age': 30} as person"
        )
        assert len(result) == 1

        await conn.close()

    @pytest.mark.asyncio
    async def test_error_handling(self):
        """Test DuckDB error handling."""
        conn = DuckDBConnection(":memory:")

        # Test syntax error
        with pytest.raises(Exception):
            await conn.execute_query("INVALID SQL SYNTAX")

        # Test table not found
        with pytest.raises(Exception):
            await conn.execute_query("SELECT * FROM nonexistent_table")

        await conn.close()


class TestDuckDBSchemaIntrospector:
    """Test DuckDB schema introspection functionality."""

    @pytest.mark.asyncio
    async def test_table_listing(self, tmp_path):
        """Test DuckDB table listing."""
        db_path = tmp_path / "schema_test.duckdb"

        # Create database with tables and views
        with duckdb.connect(str(db_path)) as db_conn:
            db_conn.execute("CREATE TABLE customers (id INTEGER, name VARCHAR)")
            db_conn.execute("CREATE TABLE orders (id INTEGER, customer_id INTEGER)")
            db_conn.execute(
                "CREATE VIEW customer_orders AS SELECT * FROM customers JOIN orders ON customers.id = orders.customer_id"
            )

        # Test introspection
        conn = DuckDBConnection(str(db_path))
        introspector = DuckDBSchemaIntrospector()

        tables = await introspector.list_tables_info(conn)

        table_names = {table["table_name"] for table in tables}
        table_types = {table["table_name"]: table["table_type"] for table in tables}

        assert "customers" in table_names
        assert "orders" in table_names
        assert "customer_orders" in table_names

        # DuckDB uses different type names
        assert table_types["customers"] == "BASE TABLE"
        assert table_types["orders"] == "BASE TABLE"
        assert table_types["customer_orders"] == "VIEW"

        await conn.close()

    @pytest.mark.asyncio
    async def test_column_introspection_with_duckdb_types(self, tmp_path):
        """Test DuckDB column introspection with DuckDB-specific types."""
        db_path = tmp_path / "types_test.duckdb"

        with duckdb.connect(str(db_path)) as db_conn:
            db_conn.execute("""
                CREATE TABLE duckdb_types (
                    id INTEGER PRIMARY KEY,
                    name VARCHAR(100) DEFAULT 'Unknown',
                    price DECIMAL(10,2) DEFAULT 0.00,
                    active BOOLEAN DEFAULT TRUE,
                    tags VARCHAR[] DEFAULT [],
                    metadata JSON,
                    created_at TIMESTAMP DEFAULT NOW(),
                    updated_at TIMESTAMPTZ DEFAULT NOW(),
                    data BLOB,
                    rating DOUBLE DEFAULT 0.0
                )
            """)

        conn = DuckDBConnection(str(db_path))
        introspector = DuckDBSchemaIntrospector()

        tables = await introspector.get_tables_info(conn)
        columns = await introspector.get_columns_info(conn, tables)

        # Organize columns for easier testing
        column_info = {
            col["column_name"]: col
            for col in columns
            if col["table_name"] == "duckdb_types"
        }

        # Check DuckDB-specific types
        assert column_info["name"]["data_type"] == "VARCHAR"
        assert column_info["price"]["data_type"] == "DECIMAL(10,2)"
        assert column_info["active"]["data_type"] == "BOOLEAN"
        assert column_info["created_at"]["data_type"] == "TIMESTAMP"
        assert column_info["rating"]["data_type"] == "DOUBLE"

        # Check defaults
        assert column_info["name"]["column_default"] is not None
        assert column_info["price"]["column_default"] is not None
        assert column_info["active"]["column_default"] is not None

        await conn.close()

    @pytest.mark.asyncio
    async def test_primary_key_detection(self, tmp_path):
        """Test DuckDB primary key detection."""
        db_path = tmp_path / "pk_test.duckdb"

        with duckdb.connect(str(db_path)) as db_conn:
            db_conn.execute(
                "CREATE TABLE single_pk (id INTEGER PRIMARY KEY, data VARCHAR)"
            )
            db_conn.execute(
                "CREATE TABLE composite_pk (user_id INTEGER, role_id INTEGER, data VARCHAR, PRIMARY KEY (user_id, role_id))"
            )

        conn = DuckDBConnection(str(db_path))
        introspector = DuckDBSchemaIntrospector()

        tables = await introspector.get_tables_info(conn)
        primary_keys = await introspector.get_primary_keys_info(conn, tables)

        pk_by_table = {}
        for pk in primary_keys:
            table_name = pk["table_name"]
            if table_name not in pk_by_table:
                pk_by_table[table_name] = []
            pk_by_table[table_name].append(pk["column_name"])

        assert "id" in pk_by_table["single_pk"]
        assert "user_id" in pk_by_table["composite_pk"]
        assert "role_id" in pk_by_table["composite_pk"]

        await conn.close()

    @pytest.mark.asyncio
    async def test_foreign_key_detection(self, tmp_path):
        """Test DuckDB foreign key detection."""
        db_path = tmp_path / "fk_test.duckdb"

        with duckdb.connect(str(db_path)) as db_conn:
            db_conn.execute(
                "CREATE TABLE departments (id INTEGER PRIMARY KEY, name VARCHAR)"
            )
            db_conn.execute("""
                CREATE TABLE employees (
                    id INTEGER PRIMARY KEY,
                    name VARCHAR,
                    dept_id INTEGER,
                    FOREIGN KEY (dept_id) REFERENCES departments(id)
                )
            """)

        conn = DuckDBConnection(str(db_path))
        introspector = DuckDBSchemaIntrospector()

        tables = await introspector.get_tables_info(conn)
        foreign_keys = await introspector.get_foreign_keys_info(conn, tables)

        # Check foreign key relationship
        fk_found = False
        for fk in foreign_keys:
            if (
                fk["table_name"] == "employees"
                and fk["column_name"] == "dept_id"
                and fk["foreign_table_name"] == "departments"
                and fk["foreign_column_name"] == "id"
            ):
                fk_found = True
                break

        assert fk_found, "Foreign key relationship not detected"
        await conn.close()

    @pytest.mark.asyncio
    async def test_index_detection(self, tmp_path):
        """Test DuckDB index detection."""
        db_path = tmp_path / "index_test.duckdb"

        with duckdb.connect(str(db_path)) as db_conn:
            db_conn.execute(
                "CREATE TABLE items (id INTEGER, name VARCHAR, category VARCHAR)"
            )
            db_conn.execute("CREATE INDEX idx_item_name ON items(name)")
            db_conn.execute("CREATE UNIQUE INDEX idx_item_id ON items(id)")

        conn = DuckDBConnection(str(db_path))
        introspector = DuckDBSchemaIntrospector()

        tables = await introspector.get_tables_info(conn)
        indexes = await introspector.get_indexes_info(conn, tables)

        index_names = [idx["index_name"] for idx in indexes]

        assert "idx_item_name" in index_names
        assert "idx_item_id" in index_names

        # Check index properties
        for idx in indexes:
            if idx["index_name"] == "idx_item_name":
                assert not idx["is_unique"]
                assert "name" in idx["column_names"]
            elif idx["index_name"] == "idx_item_id":
                assert idx["is_unique"]
                assert "id" in idx["column_names"]

        await conn.close()

    @pytest.mark.asyncio
    async def test_table_pattern_filtering(self, tmp_path):
        """Test DuckDB table pattern filtering."""
        db_path = tmp_path / "pattern_test.duckdb"

        with duckdb.connect(str(db_path)) as db_conn:
            db_conn.execute("CREATE TABLE log_events (id INTEGER)")
            db_conn.execute("CREATE TABLE log_errors (id INTEGER)")
            db_conn.execute("CREATE TABLE user_data (id INTEGER)")
            db_conn.execute("CREATE TABLE config_settings (id INTEGER)")

        conn = DuckDBConnection(str(db_path))
        introspector = DuckDBSchemaIntrospector()

        # Test pattern matching
        log_tables = await introspector.get_tables_info(conn, "log_%")
        log_table_names = {table["table_name"] for table in log_tables}

        assert "log_events" in log_table_names
        assert "log_errors" in log_table_names
        assert "user_data" not in log_table_names
        assert "config_settings" not in log_table_names

        await conn.close()

    @pytest.mark.asyncio
    async def test_schema_qualified_tables(self, tmp_path):
        """Test DuckDB schema-qualified table operations."""
        db_path = tmp_path / "schema_test.duckdb"

        with duckdb.connect(str(db_path)) as db_conn:
            # Create custom schema
            db_conn.execute("CREATE SCHEMA analytics")
            db_conn.execute("CREATE TABLE analytics.reports (id INTEGER, name VARCHAR)")
            db_conn.execute("CREATE TABLE main.users (id INTEGER, name VARCHAR)")

        conn = DuckDBConnection(str(db_path))
        introspector = DuckDBSchemaIntrospector()

        # Test schema.table pattern
        analytics_tables = await introspector.get_tables_info(conn, "analytics.%")
        analytics_table_names = {
            f"{table['table_schema']}.{table['table_name']}"
            for table in analytics_tables
        }

        assert "analytics.reports" in analytics_table_names
        assert "main.users" not in analytics_table_names

        await conn.close()

    @pytest.mark.asyncio
    async def test_duckdb_specific_features(self, tmp_path):
        """Test DuckDB-specific SQL features in introspection."""
        db_path = tmp_path / "advanced.duckdb"

        with duckdb.connect(str(db_path)) as db_conn:
            db_conn.execute("""
                CREATE TABLE advanced_features (
                    id INTEGER PRIMARY KEY,
                    json_data JSON,
                    array_data INTEGER[],
                    struct_data STRUCT(name VARCHAR, age INTEGER)
                )
            """)

        # Now test introspection
        conn = DuckDBConnection(str(db_path))
        introspector = DuckDBSchemaIntrospector()

        tables = await introspector.get_tables_info(conn)
        columns = await introspector.get_columns_info(conn, tables)

        # Check advanced data types are handled
        column_names = [col["column_name"] for col in columns]

        assert "id" in column_names
        assert "json_data" in column_names
        assert "array_data" in column_names

        await conn.close()

    @pytest.mark.asyncio
    async def test_empty_database_handling(self):
        """Test handling of empty DuckDB database."""
        conn = DuckDBConnection(":memory:")
        introspector = DuckDBSchemaIntrospector()

        tables = await introspector.list_tables_info(conn)
        assert len(tables) == 0

        columns = await introspector.get_columns_info(conn, [])
        assert len(columns) == 0

        foreign_keys = await introspector.get_foreign_keys_info(conn, [])
        assert len(foreign_keys) == 0

        await conn.close()


class _FakeDuckDBConnection:
    def __init__(self):
        self._excluded_schemas: list[str] = []

    @property
    def excluded_schemas(self) -> list[str]:
        return self._excluded_schemas

    def set_excluded_schemas(self, schemas: list[str]) -> None:
        self._excluded_schemas = schemas


def test_duckdb_excluded_defaults(monkeypatch):
    """Default exclusions should include DuckDB system schemas."""
    monkeypatch.delenv("SQLSABER_DUCKDB_EXCLUDE_SCHEMAS", raising=False)
    conn = _FakeDuckDBConnection()
    introspector = DuckDBSchemaIntrospector()

    excluded = introspector._get_excluded_schemas(conn)

    assert "information_schema" in excluded
    assert "pg_catalog" in excluded
    assert "duckdb_catalog" in excluded


def test_duckdb_excluded_env_and_connection(monkeypatch):
    """Environment and connection-level exclusions should merge and keep case."""
    monkeypatch.setenv("SQLSABER_DUCKDB_EXCLUDE_SCHEMAS", "archive,Reports")
    conn = _FakeDuckDBConnection()
    conn.set_excluded_schemas(["custom_schema", "Sales"])
    introspector = DuckDBSchemaIntrospector()

    excluded = introspector._get_excluded_schemas(conn)

    assert "archive" in excluded
    assert "Reports" in excluded
    assert "custom_schema" in excluded
    assert "Sales" in excluded
    # Ensure case-sensitive duplicates are preserved
    assert "sales" not in excluded
